from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.utils.molecule_feature_utils import one_hot_encode
from deepchem.utils.typing import RDKitMol, RDKitAtom
import numpy as np
from typing import Tuple, Any
from dataclasses import dataclass


@dataclass
class MATEncoding:
    """
    Dataclass specific to the Molecular Attention Transformer [1]_.

    This dataclass class wraps around three different matrices for a given molecule: Node Features, Adjacency Matrix, and the Distance Matrix.

    Parameters
    ----------
    node_features: np.ndarray
        Node Features matrix for the molecule. For MAT, derived from the construct_node_features_matrix function.
    adjacency_matrix: np.ndarray
        Adjacency matrix for the molecule. Derived from rdkit.Chem.rdmolops.GetAdjacencyMatrix
    distance_matrix: np.ndarray
        Distance matrix for the molecule. Derived from rdkit.Chem.rdmolops.GetDistanceMatrix

    """
    node_features: np.ndarray
    adjacency_matrix: np.ndarray
    distance_matrix: np.ndarray


class MATFeaturizer(MolecularFeaturizer):
    """
    This class is a featurizer for the Molecule Attention Transformer [1]_.
    The returned value is a numpy array which consists of molecular graph descriptions:
        - Node Features
        - Adjacency Matrix
        - Distance Matrix

    References
    ---------
    .. [1] Lukasz Maziarka et al. "Molecule Attention Transformer`<https://arxiv.org/abs/2002.08264>`"

    Examples
    --------
    >>> import deepchem as dc
    >>> feat = dc.feat.MATFeaturizer()
    >>> out = feat.featurize("CCC")

    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self):
        pass

    def construct_mol(self, mol: RDKitMol) -> RDKitMol:
        """
        Processes an input RDKitMol further to be able to extract id-specific Conformers from it using mol.GetConformer().

        Parameters
        ----------
        mol: RDKitMol
          RDKit Mol object.

        Returns
        ----------
        mol: RDKitMol
            A processed RDKitMol object which is embedded, UFF Optimized and has Hydrogen atoms removed. If the former conditions are not met and there is a value error, then 2D Coordinates are computed instead.

        """
        try:
            from rdkit.Chem import AllChem
            from rdkit import Chem
        except ModuleNotFoundError:
            pass
        try:
            mol = Chem.AddHs(mol)
            AllChem.EmbedMolecule(mol, maxAttempts=5000)
            AllChem.UFFOptimizeMolecule(mol)
            mol = Chem.RemoveHs(mol)
        except ValueError:
            AllChem.Compute2DCoords(mol)

        return mol

    def atom_features(self, atom: RDKitAtom) -> np.ndarray:
        """Deepchem already contains an atom_features function, however we are defining a new one here due to the need to handle features specific to MAT.
        Since we need new features like Atom GetNeighbors and IsInRing, and the number of features required for MAT is a fraction of what the Deepchem atom_features function computes, we can speed up computation by defining a custom function.

        Parameters
        ----------
        atom: RDKitAtom
            RDKit Atom object.

        Returns
        ----------
        ndarray
            Numpy array containing atom features.

        """
        attrib = []
        attrib += one_hot_encode(atom.GetAtomicNum(),
                                 [5, 6, 7, 8, 9, 15, 16, 17, 35, 53, 999])
        attrib += one_hot_encode(len(atom.GetNeighbors()), [0, 1, 2, 3, 4, 5])
        attrib += one_hot_encode(atom.GetTotalNumHs(), [0, 1, 2, 3, 4])

        attrib += one_hot_encode(atom.GetFormalCharge(),
                                 [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5])

        attrib.append(atom.IsInRing())
        attrib.append(atom.GetIsAromatic())

        return np.array(attrib, dtype=np.float32)

    def construct_node_features_matrix(self, mol: RDKitMol) -> np.ndarray:
        """This function constructs a matrix of atom features for all atoms in a given molecule using the atom_features function.

        Parameters
        ----------
        mol: RDKitMol
            RDKit Mol object.

        Returns
        ----------
        Atom_features: ndarray
            Numpy array containing atom features.

        """
        return np.array([self.atom_features(atom) for atom in mol.GetAtoms()])

    def _add_dummy_node(
            self, node_features: np.ndarray, adj_matrix: np.ndarray,
            dist_matrix: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Adds a single dummy node to the molecule, which is consequently reflected in the Node Features Matrix, Adjacency Matrix and the Distance Matrix.

        Parameters
        ----------
        node_features: np.ndarray
            Node Features matrix for a given molecule.
        adjacency_matrix: np.ndarray
            Adjacency matrix for a given molecule.
        distance_matrix: np.ndarray
            Distance matrix for a given molecule.

        Returns
        ----------
        Atom_features: Tuple[np.ndarray, np.ndarray, np.ndarray]
            A tuple containing three numpy arrays: node_features, adjacency_matrix, distance_matrix.

        """

        if node_features is not None:
            m = np.zeros(
                (node_features.shape[0] + 1, node_features.shape[1] + 1))
            m[1:, 1:] = node_features
            m[0, 0] = 1.0
            node_features = m

        if adj_matrix is not None:
            m = np.zeros((adj_matrix.shape[0] + 1, adj_matrix.shape[1] + 1))
            m[1:, 1:] = adj_matrix
            adj_matrix = m

        if dist_matrix is not None:
            m = np.full((dist_matrix.shape[0] + 1, dist_matrix.shape[1] + 1),
                        1e6)
            m[1:, 1:] = dist_matrix
            dist_matrix = m

        return node_features, adj_matrix, dist_matrix

    def _pad_array(self, array: np.ndarray, shape: Any) -> np.ndarray:
        """Pads an array to the desired shape.

        Parameters
        ----------
        array: np.ndarray
          Array to be padded.
        shape: int or Tuple
          Shape the array is padded to.

        Returns
        ----------
        array: np.ndarray
            Array padded to input shape.

        """
        result = np.zeros(shape=shape)
        slices = tuple(slice(s) for s in array.shape)
        result[slices] = array
        return result

    def _pad_sequence(self, sequence: np.ndarray) -> np.ndarray:
        """Pads a given sequence using the pad_array function.

        Parameters
        ----------
        sequence: np.ndarray
            Arrays in this sequence are padded to the largest shape in the sequence.

        Returns
        ----------
        array: np.ndarray
            Sequence with padded arrays.

        """
        shapes = np.stack([np.array(t.shape) for t in sequence])
        max_shape = tuple(np.max(shapes, axis=0))
        return np.stack([self._pad_array(t, shape=max_shape) for t in sequence])

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> MATEncoding:
        """
        Featurize the molecule.

        Parameters
        ----------
        datapoint: RDKitMol
            RDKit mol object.

        Returns
        -------
        MATEncoding
            A MATEncoding dataclass instance consisting of processed node_features, adjacency_matrix and distance_matrix.

        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        from rdkit import Chem

        datapoint = self.construct_mol(datapoint)

        node_features = self.construct_node_features_matrix(datapoint)
        adjacency_matrix = Chem.GetAdjacencyMatrix(datapoint)
        distance_matrix = Chem.GetDistanceMatrix(datapoint)

        node_features, adjacency_matrix, distance_matrix = self._add_dummy_node(
            node_features, adjacency_matrix, distance_matrix)

        node_features = self._pad_sequence(node_features)
        adjacency_matrix = self._pad_sequence(adjacency_matrix)
        distance_matrix = self._pad_sequence(distance_matrix)

        return MATEncoding(node_features, adjacency_matrix, distance_matrix)



import unittest
import numpy as np
from deepchem.utils import hash_utils


def random_string(length, chars=None):
    import string
    if chars is None:
        chars = list(string.ascii_letters + string.ascii_letters +
                     '()[]+-.=#@/\\')
    return ''.join(np.random.choice(chars, length))


class TestHashUtils(unittest.TestCase):

    def test_hash_ecfp(self):
        for power in (2, 16, 64):
            for _ in range(10):
                string = random_string(10)
                string_hash = hash_utils.hash_ecfp(string, power)
                self.assertIsInstance(string_hash, int)
                self.assertLess(string_hash, 2**power)
                self.assertGreaterEqual(string_hash, 0)

    def test_hash_ecfp_pair(self):
        for power in (2, 16, 64):
            for _ in range(10):
                string1 = random_string(10)
                string2 = random_string(10)
                pair_hash = hash_utils.hash_ecfp_pair((string1, string2), power)
                self.assertIsInstance(pair_hash, int)
                self.assertLess(pair_hash, 2**power)
                self.assertGreaterEqual(pair_hash, 0)

    def test_vectorize(self):
        size = 16
        feature_dict = {0: "C", 1: "CC", 2: "CCC"}
        hash_function = hash_utils.hash_ecfp
        vector = hash_utils.vectorize(hash_function, feature_dict, size)
        assert vector.shape == (size,)
        assert np.count_nonzero(vector) == len(feature_dict)



import numpy as np


def get_distribution(GC_fraction):
    from simdna.util import DiscreteDistribution
    return DiscreteDistribution({
        'A': (1 - GC_fraction) / 2,
        'C': GC_fraction / 2,
        'G': GC_fraction / 2,
        'T': (1 - GC_fraction) / 2
    })


def simple_motif_embedding(motif_name, seq_length, num_seqs, GC_fraction):
    """Simulates sequences with a motif embedded anywhere in the sequence.

    Parameters
    ----------
    motif_name : str
        encode motif name
    seq_length : int
        length of sequence
    num_seqs: int
        number of sequences
    GC_fraction : float
        GC basepair fraction in background sequence

    Returns
    -------
    sequence_arr : 1darray
        Array with sequence strings.
    embedding_arr: 1darray
        Array of embedding objects.
    """
    import simdna
    from simdna import synthetic
    if motif_name is None:
        embedders = []
    else:
        loaded_motifs = synthetic.LoadedEncodeMotifs(simdna.ENCODE_MOTIFS_PATH,
                                                     pseudocountProb=0.001)
        substring_generator = synthetic.PwmSamplerFromLoadedMotifs(
            loaded_motifs, motif_name)
        embedders = [
            synthetic.SubstringEmbedder(
                synthetic.ReverseComplementWrapper(substring_generator))
        ]
    embed_in_background = synthetic.EmbedInABackground(
        synthetic.ZeroOrderBackgroundGenerator(
            seq_length, discreteDistribution=get_distribution(GC_fraction)),
        embedders)
    generated_sequences = tuple(
        synthetic.GenerateSequenceNTimes(embed_in_background,
                                         num_seqs).generateSequences())
    sequence_arr = np.array(
        [generated_seq.seq for generated_seq in generated_sequences])
    embedding_arr = [
        generated_seq.embeddings for generated_seq in generated_sequences
    ]
    return sequence_arr, embedding_arr


def motif_density(motif_name,
                  seq_length,
                  num_seqs,
                  min_counts,
                  max_counts,
                  GC_fraction,
                  central_bp=None):
    """Returns sequences with motif density, along with embeddings array."""
    import simdna
    from simdna import synthetic
    loaded_motifs = synthetic.LoadedEncodeMotifs(simdna.ENCODE_MOTIFS_PATH,
                                                 pseudocountProb=0.001)
    substring_generator = synthetic.PwmSamplerFromLoadedMotifs(
        loaded_motifs, motif_name)
    if central_bp is not None:
        position_generator = synthetic.InsideCentralBp(central_bp)
    else:
        position_generator = synthetic.UniformPositionGenerator()
    quantity_generator = synthetic.UniformIntegerGenerator(
        min_counts, max_counts)
    embedders = [
        synthetic.RepeatedEmbedder(
            synthetic.SubstringEmbedder(
                synthetic.ReverseComplementWrapper(substring_generator),
                position_generator), quantity_generator)
    ]
    embed_in_background = synthetic.EmbedInABackground(
        synthetic.ZeroOrderBackgroundGenerator(
            seq_length, discreteDistribution=get_distribution(GC_fraction)),
        embedders)
    generated_sequences = tuple(
        synthetic.GenerateSequenceNTimes(embed_in_background,
                                         num_seqs).generateSequences())
    sequence_arr = np.array(
        [generated_seq.seq for generated_seq in generated_sequences])
    embedding_arr = [
        generated_seq.embeddings for generated_seq in generated_sequences
    ]
    return sequence_arr, embedding_arr


def simulate_single_motif_detection(motif_name, seq_length, num_pos, num_neg,
                                    GC_fraction):
    """Simulates two classes of seqeuences:
    - Positive class sequence with a motif
        embedded anywhere in the sequence
    - Negative class sequence without the motif

    Parameters
    ----------
    motif_name : str
        encode motif name
    seq_length : int
        length of sequence
    num_pos : int
        number of positive class sequences
    num_neg : int
        number of negative class sequences
    GC_fraction : float
        GC fraction in background sequence

    Returns
    -------
    sequence_arr : 1darray
        Array with sequence strings.
    y : 1darray
        Array with positive/negative class labels.
    embedding_arr: 1darray
        Array of embedding objects.
    """
    motif_sequence_arr, positive_embedding_arr = simple_motif_embedding(
        motif_name, seq_length, num_pos, GC_fraction)
    random_sequence_arr, negative_embedding_arr = simple_motif_embedding(
        None, seq_length, num_neg, GC_fraction)
    sequence_arr = np.concatenate((motif_sequence_arr, random_sequence_arr))
    y = np.array([[True]] * num_pos + [[False]] * num_neg)
    embedding_arr = positive_embedding_arr + negative_embedding_arr
    return sequence_arr, y, embedding_arr


def simulate_motif_counting(motif_name, seq_length, pos_counts, neg_counts,
                            num_pos, num_neg, GC_fraction):
    """Generates data for motif counting task.

    Parameters
    ----------
    motif_name : str
    seq_length : int
    pos_counts : list
        (min_counts, max_counts) for positive set.
    neg_counts : list
        (min_counts, max_counts) for negative set.
    num_pos : int
    num_neg : int
    GC_fraction : float

    Returns
    -------
    sequence_arr : 1darray
        Contains sequence strings.
    y : 1darray
        Contains labels.
    embedding_arr: 1darray
        Array of embedding objects.
    """
    pos_count_sequence_array, positive_embedding_arr = motif_density(
        motif_name, seq_length, num_pos, pos_counts[0], pos_counts[1],
        GC_fraction)
    neg_count_sequence_array, negative_embedding_arr = motif_density(
        motif_name, seq_length, num_pos, neg_counts[0], neg_counts[1],
        GC_fraction)
    sequence_arr = np.concatenate(
        (pos_count_sequence_array, neg_count_sequence_array))
    y = np.array([[True]] * num_pos + [[False]] * num_neg)
    embedding_arr = positive_embedding_arr + negative_embedding_arr
    return sequence_arr, y, embedding_arr


def simulate_motif_density_localization(motif_name, seq_length, center_size,
                                        min_motif_counts, max_motif_counts,
                                        num_pos, num_neg, GC_fraction):
    """Simulates two classes of seqeuences:
    - Positive class sequences with multiple motif instances in center of the sequence.
    - Negative class sequences with multiple motif instances anywhere in the sequence.

    The number of motif instances is uniformly sampled between minimum and maximum motif counts.

    Parameters
    ----------
    motif_name : str
        encode motif name
    seq_length : int
        length of sequence
    center_size : int
        length of central part of the sequence where motifs can be positioned
    min_motif_counts : int
        minimum number of motif instances
    max_motif_counts : int
        maximum number of motif instances
    num_pos : int
        number of positive class sequences
    num_neg : int
        number of negative class sequences
    GC_fraction : float
        GC fraction in background sequence

    Returns
    -------
    sequence_arr : 1darray
        Contains sequence strings.
    y : 1darray
        Contains labels.
    embedding_arr: 1darray
        Array of embedding objects.
    """
    localized_density_sequence_array, positive_embedding_arr = motif_density(
        motif_name, seq_length, num_pos, min_motif_counts, max_motif_counts,
        GC_fraction, center_size)
    unlocalized_density_sequence_array, negative_embedding_arr = motif_density(
        motif_name, seq_length, num_neg, min_motif_counts, max_motif_counts,
        GC_fraction)
    sequence_arr = np.concatenate(
        (localized_density_sequence_array, unlocalized_density_sequence_array))
    y = np.array([[True]] * num_pos + [[False]] * num_neg)
    embedding_arr = positive_embedding_arr + negative_embedding_arr
    return sequence_arr, y, embedding_arr


def simulate_multi_motif_embedding(motif_names, seq_length, min_num_motifs,
                                   max_num_motifs, num_seqs, GC_fraction):
    """Generates data for multi motif recognition task.

    Parameters
    ----------
    motif_names : list
        List of strings.
    seq_length : int
    min_num_motifs : int
    max_num_motifs : int
    num_seqs : int
    GC_fraction : float

    Returns
    -------
    sequence_arr : 1darray
        Contains sequence strings.
    y : ndarray
        Contains labels for each motif.
    embedding_arr: 1darray
        Array of embedding objects.
    """

    import simdna
    from simdna import synthetic
    loaded_motifs = synthetic.LoadedEncodeMotifs(simdna.ENCODE_MOTIFS_PATH,
                                                 pseudocountProb=0.001)

    def get_embedder(motif_name):
        substring_generator = synthetic.PwmSamplerFromLoadedMotifs(
            loaded_motifs, motif_name)
        return synthetic.SubstringEmbedder(
            synthetic.ReverseComplementWrapper(substring_generator),
            name=motif_name)

    embedders = [get_embedder(motif_name) for motif_name in motif_names]
    quantity_generator = synthetic.UniformIntegerGenerator(
        min_num_motifs, max_num_motifs)
    combined_embedder = [
        synthetic.RandomSubsetOfEmbedders(quantity_generator, embedders)
    ]
    embed_in_background = synthetic.EmbedInABackground(
        synthetic.ZeroOrderBackgroundGenerator(
            seq_length, discreteDistribution=get_distribution(GC_fraction)),
        combined_embedder)
    generated_sequences = tuple(
        synthetic.GenerateSequenceNTimes(embed_in_background,
                                         num_seqs).generateSequences())
    sequence_arr = np.array(
        [generated_seq.seq for generated_seq in generated_sequences])
    label_generator = synthetic.IsInTraceLabelGenerator(np.asarray(motif_names))
    y = np.array([
        label_generator.generateLabels(generated_seq)
        for generated_seq in generated_sequences
    ],
                 dtype=bool)
    embedding_arr = [
        generated_seq.embeddings for generated_seq in generated_sequences
    ]
    return sequence_arr, y, embedding_arr


def simulate_differential_accessibility(pos_motif_names, neg_motif_names,
                                        seq_length, min_num_motifs,
                                        max_num_motifs, num_pos, num_neg,
                                        GC_fraction):
    """Generates data for differential accessibility task.

    Parameters
    ----------
    pos_motif_names : list
        List of strings.
    neg_motif_names : list
        List of strings.
    seq_length : int
    min_num_motifs : int
    max_num_motifs : int
    num_pos : int
    num_neg : int
    GC_fraction : float

    Returns
    -------
    sequence_arr : 1darray
        Contains sequence strings.
    y : 1darray
        Contains labels.
    embedding_arr: 1darray
        Array of embedding objects.
    """
    pos_motif_sequence_arr, _, positive_embedding_arr = simulate_multi_motif_embedding(
        pos_motif_names, seq_length, min_num_motifs, max_num_motifs, num_pos,
        GC_fraction)
    neg_motif_sequence_arr, _, negative_embedding_arr = simulate_multi_motif_embedding(
        neg_motif_names, seq_length, min_num_motifs, max_num_motifs, num_neg,
        GC_fraction)
    sequence_arr = np.concatenate(
        (pos_motif_sequence_arr, neg_motif_sequence_arr))
    y = np.array([[True]] * num_pos + [[False]] * num_neg)
    embedding_arr = positive_embedding_arr + negative_embedding_arr
    return sequence_arr, y, embedding_arr


def simulate_heterodimer_grammar(motif1, motif2, seq_length, min_spacing,
                                 max_spacing, num_pos, num_neg, GC_fraction):
    """Simulates two classes of sequences with motif1 and motif2:
        - Positive class sequences with motif1 and motif2 positioned min_spacing and max_spacing
        - Negative class sequences with independent motif1 and motif2 positioned anywhere in the sequence, not as a heterodimer grammar

    Parameters
    ----------
    seq_length : int, length of sequence
    GC_fraction : float, GC fraction in background sequence
    num_pos : int, number of positive class sequences
    num_neg : int, number of negatice class sequences
    motif1 : str, encode motif name
    motif2 : str, encode motif name
    min_spacing : int, minimum inter motif spacing
    max_spacing : int, maximum inter motif spacing

    Returns
    -------
    sequence_arr : 1darray
        Array with sequence strings.
    y : 1darray
        Array with positive/negative class labels.
    embedding_arr: list
        List of embedding objects.
    """
    import simdna
    from simdna import synthetic
    loaded_motifs = synthetic.LoadedEncodeMotifs(simdna.ENCODE_MOTIFS_PATH,
                                                 pseudocountProb=0.001)
    motif1_generator = synthetic.ReverseComplementWrapper(
        synthetic.PwmSamplerFromLoadedMotifs(loaded_motifs, motif1))
    motif2_generator = synthetic.ReverseComplementWrapper(
        synthetic.PwmSamplerFromLoadedMotifs(loaded_motifs, motif2))
    separation_generator = synthetic.UniformIntegerGenerator(
        min_spacing, max_spacing)
    embedder = synthetic.EmbeddableEmbedder(
        synthetic.PairEmbeddableGenerator(motif1_generator, motif2_generator,
                                          separation_generator))
    embed_in_background = synthetic.EmbedInABackground(
        synthetic.ZeroOrderBackgroundGenerator(
            seq_length, discreteDistribution=get_distribution(GC_fraction)),
        [embedder])
    generated_sequences = tuple(
        synthetic.GenerateSequenceNTimes(embed_in_background,
                                         num_pos).generateSequences())
    grammar_sequence_arr = np.array(
        [generated_seq.seq for generated_seq in generated_sequences])
    positive_embedding_arr = [
        generated_seq.embeddings for generated_seq in generated_sequences
    ]
    nongrammar_sequence_arr, _, negative_embedding_arr = simulate_multi_motif_embedding(
        [motif1, motif2], seq_length, 2, 2, num_neg, GC_fraction)
    sequence_arr = np.concatenate(
        (grammar_sequence_arr, nongrammar_sequence_arr))
    y = np.array([[True]] * num_pos + [[False]] * num_neg)
    embedding_arr = positive_embedding_arr + negative_embedding_arr
    return sequence_arr, y, embedding_arr



from deepchem.models.torch_models import TorchModel
from deepchem.models.losses import HingeLoss
import torch.nn as nn
import torch.nn.functional as F


class ScScore(nn.Module):
    """
    Builder class for ScScore Model.

    The SCScore model is a neural network model based on the work of Coley et al. [1]_ that predicts the synthetic complexity score (SCScore) of molecules and correlates it with the expected number of reaction steps required to produce the given target molecule.
    It is trained on a dataset of over 12 million reactions from the Reaxys database to impose a pairwise inequality constraint enforcing that on average the products of published chemical reactions should be more synthetically complex than their corresponding reactants.
    The learned metric (SCScore) exhibits highly desirable nonlinear behavior, particularly in recognizing increases in synthetic complexity throughout a number of linear synthetic routes.
    The SCScore model can accurately predict the synthetic complexity of a variety of molecules, including both drug-like and natural product molecules.
    SCScore has the potential to be a valuable tool for chemists who are working on drug discovery and other areas of chemistry.

    The learned metric (SCScore) exhibits highly desirable nonlinear behavior, particularly in recognizing increases in synthetic complexity throughout a number of linear synthetic routes.

    Our model uses hingeloss instead of the shifted relu loss as in the supplementary material [2]_ provided by the author.
    This could cause differentiation issues with compounds that are "close" to each other in "complexity".

    The default values for the model are the same as the ones used in the original paper [1]_.

    Examples
    --------
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models import ScScoreModel
    >>> # preparing dataset
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> labels = [0., 1.]
    >>> featurizer = dc.feat.CircularFingerprint(size=1024, radius=2, chiral=True)
    >>> X = featurizer.featurize(smiles)
    >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
    >>> # training model
    >>> model = ScScoreModel(n_features=1024, layer_sizes=[300, 300, 300, 300, 300],
    ...                      dropout=0.2, score_scale=5)
    >>> loss = model.fit(dataset, nb_epoch=5)

    References
    ----------
    .. [1] Coley, C. W., Rogers, L., Green, W., & Jensen, K. F. (2018). "SCScore: Synthetic Complexity Learned from a Reaction Corpus". Journal of Chemical Information and Modeling, 58(2), 252-261. https://doi.org/10.1021/acs.jcim.7b00622

    .. [2] Coley, C. W., Rogers, L., Green, W., & Jensen, K. F. (2018). Supplementary material to "SCScore: Synthetic Complexity Learned from a Reaction Corpus". Journal of Chemical Information and Modeling, 58(2), 252-261. https://github.com/connorcoley/scscore
    """

    def __init__(self,
                 n_features: int = 1024,
                 layer_sizes: list = [300, 300, 300, 300, 300],
                 dropout: float = 0.0,
                 score_scale: int = 5):
        """
        Parameters
        ----------
        n_features: int (default 1024)
            Number of features per molecule.
        layer_sizes: list of int (default [300, 300, 300, 300, 300])
            Size of each hidden layer, length of the list is the number of hidden layers.
        dropouts: float (default 0.0)
            Droupout to apply to each hidden layer.
        score_scale: int (default 5)
            Scale of the output score.
        """
        super(ScScore, self).__init__()
        self.n_features = n_features
        self.layer_sizes = layer_sizes
        self.dropout = dropout
        self.score_scale = score_scale

        input_size = self.layer_sizes[0]
        self.input_layer = nn.Linear(self.n_features, input_size)

        self.hidden_layers = nn.ModuleList()
        for layer_size in self.layer_sizes[1:]:
            self.hidden_layers.append(nn.Linear(input_size, layer_size))
            input_size = layer_size

        self.output_layer = nn.Linear(self.layer_sizes[-1], 1)

    def forward(self, inputs):
        """
        Parameters
        ----------
        inputs: list of Tensor
            data inputs for the model

        Returns
        -------
        output: Tensor
            Synthetic complexity score of the input molecule.
        """

        x = F.relu(self.input_layer(inputs))
        if self.dropout > 0.0:
            x = F.dropout(x, p=self.dropout)

        for hidden_layer in self.hidden_layers:
            x = F.relu(hidden_layer(x))

            if self.dropout > 0.0:
                x = F.dropout(x, p=self.dropout)

        output = F.sigmoid(self.output_layer(x))

        output = 1 + (self.score_scale - 1) * output
        return output


class ScScoreModel(TorchModel):
    """
    The SCScore model is a neural network model based on the work of Coley et al. [1]_ that predicts the synthetic complexity score (SCScore) of molecules and correlates it with the expected number of reaction steps required to produce the given target molecule.
    It is trained on a dataset of over 12 million reactions from the Reaxys database to impose a pairwise inequality constraint enforcing that on average the products of published chemical reactions should be more synthetically complex than their corresponding reactants.
    The learned metric (SCScore) exhibits highly desirable nonlinear behavior, particularly in recognizing increases in synthetic complexity throughout a number of linear synthetic routes.
    The SCScore model can accurately predict the synthetic complexity of a variety of molecules, including both drug-like and natural product molecules.
    SCScore has the potential to be a valuable tool for chemists who are working on drug discovery and other areas of chemistry.

    The learned metric (SCScore) exhibits highly desirable nonlinear behavior, particularly in recognizing increases in synthetic complexity throughout a number of linear synthetic routes.

    Our model uses hingeloss instead of the shifted relu loss as in the supplementary material [2]_ provided by the author.
    This could cause differentiation issues with compounds that are "close" to each other in "complexity".

    The default values for the model are the same as the ones used in the original paper [1]_.

    Examples
    --------
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models import ScScoreModel
    >>> # preparing dataset
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> labels = [0., 1.]
    >>> featurizer = dc.feat.CircularFingerprint(size=1024, radius=2, chiral=True)
    >>> X = featurizer.featurize(smiles)
    >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
    >>> # training model
    >>> model = ScScoreModel(n_features=1024, layer_sizes=[300, 300, 300, 300, 300],
    ...                      dropout=0.2, score_scale=5)
    >>> loss = model.fit(dataset, nb_epoch=5)

    References
    ----------
    .. [1] Coley, C. W., Rogers, L., Green, W., & Jensen, K. F. (2018). "SCScore: Synthetic Complexity Learned from a Reaction Corpus". Journal of Chemical Information and Modeling, 58(2), 252-261. https://doi.org/10.1021/acs.jcim.7b00622

    .. [2] Coley, C. W., Rogers, L., Green, W., & Jensen, K. F. (2018). Supplementary material to "SCScore: Synthetic Complexity Learned from a Reaction Corpus". Journal of Chemical Information and Modeling, 58(2), 252-261. https://github.com/connorcoley/scscore
    """

    def __init__(self,
                 n_features: int = 1024,
                 layer_sizes: list = [300, 300, 300, 300, 300],
                 dropout: float = 0.0,
                 score_scale: int = 5,
                 **kwargs):
        """
        Parameters
        ----------
        n_features: int (default 1024)
            Number of features per molecule.
        layer_sizes: list of int (default [300, 300, 300, 300, 300])
            Size of each hidden layer, length of the list is the number of hidden layers.
        dropouts: float (default 0.0)
            Droupout to apply to each hidden layer.
        score_scale: int (default 5)
            Max output score, range of output = (1, score_scale).
        """

        if dropout < 0.0 or dropout > 1.0:
            raise ValueError(
                f"Dropout must be between 0.0 and 1.0, but got {dropout}")

        if score_scale < 1:
            raise ValueError(
                f"Score scale must be greater than 1, but got {score_scale}")

        model = ScScore(n_features=n_features,
                        layer_sizes=layer_sizes,
                        dropout=dropout,
                        score_scale=score_scale)

        if 'loss' not in kwargs:
            kwargs['loss'] = HingeLoss()

        super(ScScoreModel, self).__init__(model, **kwargs)



from typing import TYPE_CHECKING
from deepchem.feat.base_classes import Featurizer

if TYPE_CHECKING:
    import transformers


class HuggingFaceFeaturizer(Featurizer):
    """Wrapper class that wraps HuggingFace tokenizers as DeepChem featurizers

    The `HuggingFaceFeaturizer` wrapper provides a wrapper
    around Hugging Face tokenizers allowing them to be used as DeepChem
    featurizers. This might be useful in scenarios where user needs to use
    a hugging face tokenizer when loading a dataset.

    Example
    -------
    >>> from deepchem.feat import HuggingFaceFeaturizer
    >>> from transformers import RobertaTokenizerFast
    >>> hf_tokenizer = RobertaTokenizerFast.from_pretrained("seyonec/PubChem10M_SMILES_BPE_60k")
    >>> featurizer = HuggingFaceFeaturizer(tokenizer=hf_tokenizer)
    >>> result = featurizer.featurize(['CC(=O)C'])
    """

    def __init__(
        self,
        tokenizer: 'transformers.tokenization_utils_fast.PreTrainedTokenizerFast'
    ):
        """Initializes a tokenizer wrapper

        Parameters
        ----------
        tokenizer: transformers.tokenization_utils_fast.PreTrainedTokenizerFast
            The tokenizer to use for featurization
        """
        self.tokenizer = tokenizer

    def _featurize(self, datapoint):
        """Featurizes a single datapoint using the tokenizer"""
        return self.tokenizer(datapoint).data



import pytest
try:
    import torch
except:
    print("torch not available")
try:
    import pylibxc
except:
    print("pylibxc not available")

import numpy as np
from typing import Union


@pytest.mark.torch
def test_base_orb_params():
    from deepchem.utils.dft_utils import BaseOrbParams

    class MyOrbParams(BaseOrbParams):

        @staticmethod
        def params2orb(params, coeffs, with_penalty):
            return params, coeffs

        @staticmethod
        def orb2params(orb):
            return orb, torch.tensor([0], dtype=orb.dtype, device=orb.device)

    params = torch.randn(3, 4, 5)
    coeffs = torch.randn(3, 4, 5)
    with_penalty = 0.1
    orb, penalty = MyOrbParams.params2orb(params, coeffs, with_penalty)
    params2, coeffs2 = MyOrbParams.orb2params(orb)
    assert torch.allclose(params, params2)


@pytest.mark.torch
def test_qr_orb_params():
    from deepchem.utils.dft_utils import QROrbParams
    params = torch.randn(3, 3)
    coeffs = torch.randn(4, 3)
    with_penalty = 0.1
    orb, penalty = QROrbParams.params2orb(params, coeffs, with_penalty)
    params2, coeffs2 = QROrbParams.orb2params(orb)
    assert torch.allclose(orb, params2)


@pytest.mark.torch
def test_mat_exp_orb_params():
    from deepchem.utils.dft_utils import MatExpOrbParams
    params = torch.randn(3, 3)
    coeffs = torch.randn(4, 3)
    orb = MatExpOrbParams.params2orb(params, coeffs)[0]
    params2, coeffs2 = MatExpOrbParams.orb2params(orb)
    assert coeffs2.shape == orb.shape


@pytest.mark.torch
def test_lattice():
    """Test lattice object.
    Comparing it's output with it's original implementation in dqc.
    """
    from deepchem.utils.dft_utils import Lattice
    a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
    lattice = Lattice(a)

    assert torch.allclose(lattice.lattice_vectors(), a)
    assert torch.allclose(lattice.recip_vectors(),
                          torch.inverse(a.transpose(-2, -1)) * (2 * np.pi))
    assert torch.allclose(lattice.volume(), torch.det(a))

    assert torch.allclose(
        lattice.get_lattice_ls(1.0),
        torch.tensor([[0., 0., -1.], [0., -1., 0.], [-1., 0., 0.], [0., 0., 0.],
                      [1., 0., 0.], [0., 1., 0.], [0., 0., 1.]]))
    assert torch.allclose(
        lattice.get_gvgrids(6.0)[0],
        torch.tensor([[0.0000, 0.0000, -6.2832], [0.0000, -6.2832, 0.0000],
                      [-6.2832, 0.0000, 0.0000], [0.0000, 0.0000, 0.0000],
                      [6.2832, 0.0000, 0.0000], [0.0000, 6.2832, 0.0000],
                      [0.0000, 0.0000, 6.2832]]))
    assert torch.allclose(
        lattice.get_gvgrids(6.0)[1],
        torch.tensor([1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000]))
    assert lattice.estimate_ewald_eta(1e-5) == 1.8


@pytest.mark.torch
def test_parse_moldesc():
    """Tests Moldesc Parser."""
    from deepchem.utils.dft_utils import parse_moldesc
    system = {
        'type': 'mol',
        'kwargs': {
            'moldesc': 'H 0.86625 0 0; F -0.86625 0 0',
            'basis': '6-311++G(3df,3pd)'
        }
    }
    atomzs, atomposs = parse_moldesc(system["kwargs"]["moldesc"])
    assert torch.allclose(atomzs, torch.tensor([1., 9.], dtype=torch.float64))
    assert torch.allclose(
        atomposs,
        torch.tensor(
            [[0.86625, 0.00000, 0.00000], [-0.86625, 0.00000, 0.00000]],
            dtype=torch.float64))
    system2 = (['H', 'F'], torch.tensor([[0.86625, 0, 0], [-0.86625, 0, 0]]))
    atomzs2, atomposs2 = parse_moldesc(system2)
    assert torch.allclose(atomzs2, torch.tensor([1., 9.], dtype=torch.float64))
    assert torch.allclose(
        atomposs2,
        torch.tensor(
            [[0.86625, 0.00000, 0.00000], [-0.86625, 0.00000, 0.00000]],
            dtype=torch.float64))


@pytest.mark.torch
def test_spin_param():
    """Test SpinParam object."""
    from deepchem.utils.dft_utils import SpinParam
    dens_u = torch.ones(1)
    dens_d = torch.zeros(1)
    sp = SpinParam(u=dens_u, d=dens_d)

    assert torch.allclose(sp.u, dens_u)
    assert torch.allclose(sp.d, dens_d)
    assert torch.allclose(sp.sum(), torch.tensor([1.]))
    assert torch.allclose(sp.reduce(torch.multiply), torch.tensor([0.]))


@pytest.mark.torch
def test_val_grad():
    """Test ValGrad object."""
    from deepchem.utils.dft_utils import ValGrad
    dens = torch.ones(1)
    grad = torch.zeros(1)
    lapl = torch.ones(1)
    kin = torch.ones(1)
    vg = ValGrad(value=dens, grad=grad, lapl=lapl, kin=kin)

    assert torch.allclose(vg.value, dens)
    assert torch.allclose(vg.grad, grad)
    assert torch.allclose(vg.lapl, lapl)
    assert torch.allclose(vg.kin, kin)

    vg2 = vg + vg
    assert torch.allclose(vg2.value, torch.tensor([2.]))
    assert torch.allclose(vg2.grad, torch.tensor([0.]))
    assert torch.allclose(vg2.lapl, torch.tensor([2.]))
    assert torch.allclose(vg2.kin, torch.tensor([2.]))

    vg5 = vg * 5
    assert torch.allclose(vg5.value, torch.tensor([5.]))
    assert torch.allclose(vg5.grad, torch.tensor([0.]))
    assert torch.allclose(vg5.lapl, torch.tensor([5.]))
    assert torch.allclose(vg5.kin, torch.tensor([5.]))


@pytest.mark.torch
def test_base_xc():
    """Test BaseXC."""
    from deepchem.utils.dft_utils import ValGrad, SpinParam
    from deepchem.utils.dft_utils import BaseXC

    class MyXC(BaseXC):

        @property
        def family(self) -> int:
            return 1

        def get_edensityxc(
                self, densinfo: Union[ValGrad,
                                      SpinParam[ValGrad]]) -> torch.Tensor:
            if isinstance(densinfo, ValGrad):
                return densinfo.value.pow(2)
            else:
                return densinfo.u.value.pow(2) + densinfo.d.value.pow(2)

        def get_vxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]
        ) -> Union[ValGrad, SpinParam[ValGrad]]:
            if isinstance(densinfo, ValGrad):
                return ValGrad(value=2 * densinfo.value)
            else:
                return SpinParam(u=ValGrad(value=2 * densinfo.u.value),
                                 d=ValGrad(value=2 * densinfo.d.value))

    xc = MyXC()
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc.get_edensityxc(densinfo_v),
                          torch.tensor([1., 4., 9.]))
    assert xc.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([2., 4., 6.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc.get_edensityxc(densinfo_s),
                          torch.tensor([17., 29., 45.]))
    assert xc.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([2., 4., 6.])),
        d=ValGrad(value=torch.tensor([8., 10., 12.])))


@pytest.mark.torch
def test_addbase_xc():
    """Test AddBaseXC."""
    from deepchem.utils.dft_utils import ValGrad, SpinParam
    from deepchem.utils.dft_utils import BaseXC, AddBaseXC

    class MyXC(BaseXC):

        @property
        def family(self) -> int:
            return 1

        def get_edensityxc(
                self, densinfo: Union[ValGrad,
                                      SpinParam[ValGrad]]) -> torch.Tensor:
            if isinstance(densinfo, ValGrad):
                return densinfo.value.pow(2)
            else:
                return densinfo.u.value.pow(2) + densinfo.d.value.pow(2)

        def get_vxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]
        ) -> Union[ValGrad, SpinParam[ValGrad]]:
            if isinstance(densinfo, ValGrad):
                return ValGrad(value=2 * densinfo.value)
            else:
                return SpinParam(u=ValGrad(value=2 * densinfo.u.value),
                                 d=ValGrad(value=2 * densinfo.d.value))

    xc1 = MyXC()
    xc = xc1 + xc1
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))

    xc2 = AddBaseXC(xc1, xc1)
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc2.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc2.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc2.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc2.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))


@pytest.mark.torch
def test_mulbase_xc():
    """Test AddBaseXC."""
    from deepchem.utils.dft_utils import ValGrad, SpinParam
    from deepchem.utils.dft_utils import BaseXC, MulBaseXC

    class MyXC(BaseXC):

        @property
        def family(self) -> int:
            return 1

        def get_edensityxc(
                self, densinfo: Union[ValGrad,
                                      SpinParam[ValGrad]]) -> torch.Tensor:
            if isinstance(densinfo, ValGrad):
                return densinfo.value.pow(2)
            else:
                return densinfo.u.value.pow(2) + densinfo.d.value.pow(2)

        def get_vxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]
        ) -> Union[ValGrad, SpinParam[ValGrad]]:
            if isinstance(densinfo, ValGrad):
                return ValGrad(value=2 * densinfo.value)
            else:
                return SpinParam(u=ValGrad(value=2 * densinfo.u.value),
                                 d=ValGrad(value=2 * densinfo.d.value))

    xc1 = MyXC()
    xc = xc1 * 2
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))

    xc2 = MulBaseXC(xc1, 2)
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc2.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc2.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc2.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc2.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))


@pytest.mark.torch
def test_base_grid():
    import torch
    from deepchem.utils.dft_utils import BaseGrid

    class Grid(BaseGrid):

        def __init__(self):
            super(Grid, self).__init__()
            self.ngrid = 10
            self.ndim = 3
            self.dvolume = torch.ones(self.ngrid)
            self.rgrid = torch.ones((self.ngrid, self.ndim))

        def get_dvolume(self):
            return self.dvolume

        def get_rgrid(self):
            return self.rgrid

    grid = Grid()
    assert torch.allclose(grid.get_dvolume(), torch.ones(10))
    assert torch.allclose(grid.get_rgrid(), torch.ones((10, 3)))


@pytest.mark.torch
def test_base_df():
    """Test BaseDF. Checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils import BaseDF

    class MyDF(BaseDF):

        def __init__(self):
            super(MyDF, self).__init__()

        def get_j2c(self):
            return torch.ones((3, 3))

        def get_j3c(self):
            return torch.ones((3, 3, 3))

    df = MyDF()
    assert torch.allclose(df.get_j2c(), torch.ones((3, 3)))
    assert torch.allclose(df.get_j3c(), torch.ones((3, 3, 3)))


@pytest.mark.torch
def test_base_hamilton():
    """Test BaseHamilton. Checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils import BaseHamilton

    class MyHamilton(BaseHamilton):

        def __init__(self):
            self._nao = 2
            self._kpts = torch.tensor([[0.0, 0.0, 0.0]])
            self._df = None

        @property
        def nao(self):
            return self._nao

        @property
        def kpts(self):
            return self._kpts

        @property
        def df(self):
            return self._df

        def build(self):
            return self

        def get_nuclattr(self):
            return torch.ones((1, 1, self.nao, self.nao))

    ham = MyHamilton()
    hamilton = ham.build()
    assert torch.allclose(hamilton.get_nuclattr(),
                          torch.tensor([[[[1., 1.], [1., 1.]]]]))


@pytest.mark.torch
def test_cgto_basis():
    from deepchem.utils.dft_utils import CGTOBasis
    alphas = torch.ones(1)
    coeffs = torch.ones(1)
    cgto = CGTOBasis(angmom=0, alphas=alphas, coeffs=coeffs)
    cgto.wfnormalize_()

    if cgto.normalized is True:
        assert True
    else:
        assert False


@pytest.mark.torch
def test_atom_cgto_basis():
    from deepchem.utils.dft_utils import AtomCGTOBasis, CGTOBasis
    alphas = torch.ones(1)
    coeffs = torch.ones(1)
    cgto = CGTOBasis(angmom=0, alphas=alphas, coeffs=coeffs)
    atomcgto = AtomCGTOBasis(atomz=1, bases=[cgto], pos=[0.0, 0.0, 0.0])
    assert atomcgto.bases[0] == cgto


@pytest.mark.torch
def test_base_system():
    """Test BaseSystem. Checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils import BaseSystem, BaseHamilton, BaseGrid

    class MySystem(BaseSystem):

        def __init__(self):
            self.hamiltonian = BaseHamilton()
            self.grid = BaseGrid()

        def get_hamiltonian(self):
            return self.hamiltonian

        def get_grid(self):
            return self.grid

        def requires_grid(self):
            return True

    system = MySystem()
    assert system.requires_grid()


@pytest.mark.torch
def test_radial_grid():
    from deepchem.utils.dft_utils import RadialGrid
    grid = RadialGrid(4, grid_integrator="chebyshev", grid_transform="logm3")
    assert grid.get_rgrid().shape == torch.Size([4, 1])
    assert grid.get_dvolume().shape == torch.Size([4])


@pytest.mark.torch
def test_get_xw_integration():
    from deepchem.utils.dft_utils import get_xw_integration
    x, w = get_xw_integration(4, "chebyshev")
    assert x.shape == (4,)
    assert w.shape == torch.Size([4])


@pytest.mark.torch
def test_sliced_radial_grid():
    from deepchem.utils.dft_utils import RadialGrid, SlicedRadialGrid
    grid = RadialGrid(4)
    sliced_grid = SlicedRadialGrid(grid, 2)
    assert sliced_grid.get_rgrid().shape == torch.Size([1])


@pytest.mark.torch
def test_de2_transform():
    from deepchem.utils.dft_utils import DE2Transformation
    x = torch.linspace(-1, 1, 100)
    r = DE2Transformation().x2r(x)
    assert r.shape == torch.Size([100])
    drdx = DE2Transformation().get_drdx(x)
    assert drdx.shape == torch.Size([100])


@pytest.mark.torch
def test_logm3_transform():
    from deepchem.utils.dft_utils import LogM3Transformation
    x = torch.linspace(-1, 1, 100)
    r = LogM3Transformation().x2r(x)
    assert r.shape == torch.Size([100])
    drdx = LogM3Transformation().get_drdx(x)
    assert drdx.shape == torch.Size([100])


@pytest.mark.torch
def test_treutlerm4_transform():
    from deepchem.utils.dft_utils import TreutlerM4Transformation
    x = torch.linspace(-1, 1, 100)
    r = TreutlerM4Transformation().x2r(x)
    assert r.shape == torch.Size([100])
    drdx = TreutlerM4Transformation().get_drdx(x)
    assert drdx.shape == torch.Size([100])


@pytest.mark.torch
def test_get_grid_transform():
    from deepchem.utils.dft_utils import get_grid_transform
    transform = get_grid_transform("logm3")
    transform.x2r(torch.tensor([0.5])) == torch.tensor([2.])


@pytest.mark.torch
def test_SCF_QCCalc():
    from deepchem.utils.dft_utils import SCF_QCCalc, BaseSCFEngine, SpinParam

    # Define the engine
    class engine(BaseSCFEngine):

        def polarised():
            return False

        def dm2energy(
                self, dm: Union[torch.Tensor,
                                SpinParam[torch.Tensor]]) -> torch.Tensor:
            if isinstance(dm, SpinParam):
                return dm.u + dm.d * 1.1
            return dm * 1.1

    myEngine = engine()
    a = SCF_QCCalc(myEngine)
    assert torch.allclose(a.dm2energy(torch.tensor([1.1])),
                          torch.tensor([1.2100]))


@pytest.mark.torch
def test_BaseSCFEngine():
    from deepchem.utils.dft_utils import BaseSCFEngine, SpinParam

    class engine(BaseSCFEngine):

        def polarised():
            return False

        def dm2energy(
                self, dm: Union[torch.Tensor,
                                SpinParam[torch.Tensor]]) -> torch.Tensor:
            if isinstance(dm, SpinParam):
                return dm.u + dm.d * 1.1
            return dm * 1.1

    myEngine = engine()
    assert myEngine.dm2energy(torch.tensor(1.2)) == torch.tensor(1.32)


@pytest.mark.torch
def test_hf_engine():
    """Tests HFEngine and methods of its parent class BaseSCFEngine."""
    from deepchem.utils.dft_utils import (BaseHamilton, BaseSystem, BaseGrid,
                                          SpinParam, HFEngine)
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

        def _rmv(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _mm(self, x):
            return torch.matmul(self.param, x)

        def _rmm(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _fullmatrix(self):
            return self.param

    class MyHamilton(BaseHamilton):

        def __init__(self):
            self._nao = 2
            self._kpts = torch.tensor([[0.0, 0.0, 0.0]])
            self._df = None

        @property
        def nao(self):
            return self._nao

        @property
        def kpts(self):
            return self._kpts

        @property
        def df(self):
            return self._df

        def build(self):
            return self

        def get_nuclattr(self):
            return torch.ones((1, 1, self.nao, self.nao))

        def get_e_elrep(self, dmtot):
            return 2 * dmtot

        def get_e_exchange(self, dm):
            return 3 * dm

        def get_e_hcore(self, dm):
            return 4.0 * dm

        def get_elrep(self, dmtot):
            return MyLinOp((self.nao + 1, self.nao + 1))

        def get_exchange(self, dm):
            return MyLinOp((self.nao + 1, self.nao + 1))

        def get_kinnucl(self):
            linop = MyLinOp((self.nao + 1, self.nao + 1))
            return linop

        def ao_orb2dm(self, orb: torch.Tensor,
                      orb_weight: torch.Tensor) -> torch.Tensor:
            return orb * orb_weight

    ham = MyHamilton()

    class MySystem(BaseSystem):

        def __init__(self):
            self.hamiltonian = ham
            self.grid = BaseGrid()

        def get_hamiltonian(self):
            return self.hamiltonian

        def get_grid(self):
            return self.grid

        def requires_grid(self):
            return True

        def get_orbweight(
            self,
            polarized: bool = False
        ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
            return SpinParam(torch.tensor([1.0]), torch.tensor([2.0]))

        def get_nuclei_energy(self):
            return 10.0

    system = MySystem()
    engine = HFEngine(system)
    engine.set_eigen_options(eigen_options={"method": "exacteig"})

    assert engine.dm2energy(torch.tensor([2])) == torch.tensor([28.0])
    assert engine.dm2scp(torch.tensor([2])).shape == torch.Size([3, 3])
    assert engine.scp2dm(torch.rand((2, 2, 2))).u.shape == torch.Size([2, 1])


@pytest.mark.torch
def test_CalcLDALibXCUnpol():
    from deepchem.utils.dft_utils import CalcLDALibXCUnpol
    libxcfcn = pylibxc.LibXCFunctional("lda_x", "unpolarized")
    rho = torch.tensor([0.1, 0.2, 0.3])
    res = CalcLDALibXCUnpol.apply(rho, 0, libxcfcn)[0]
    assert torch.allclose(
        res, torch.tensor([[-0.0343, -0.0864, -0.1483]], dtype=torch.float64),
        0.001)


@pytest.mark.torch
def test_CalcLDALibXCPol():
    from deepchem.utils.dft_utils import CalcLDALibXCPol
    libxcfcn = pylibxc.LibXCFunctional("lda_x", "polarized")
    rho_u = torch.tensor([0.1, 0.2, 0.3])
    rho_d = torch.tensor([0.1, 0.2, 0.3])
    res = CalcLDALibXCPol.apply(rho_u, rho_d, 0, libxcfcn)[0]
    assert torch.allclose(
        res, torch.tensor([[-0.0864, -0.2177, -0.3738]], dtype=torch.float64),
        0.001)


@pytest.mark.torch
def test_CalcGGALibXCUnpol():
    from deepchem.utils.dft_utils import CalcGGALibXCUnpol
    libxcfcn = pylibxc.LibXCFunctional("gga_c_pbe", "unpolarized")
    rho = torch.tensor([0.1, 0.2, 0.3])
    sigma = torch.tensor([0.1, 0.2, 0.3])
    res = CalcGGALibXCUnpol.apply(rho, sigma, 0, libxcfcn)[0]
    assert torch.allclose(
        res, torch.tensor([[-0.0016, -0.0070, -0.0137]], dtype=torch.float64),
        0.0001, 0.0001)


@pytest.mark.torch
def test_CalcGGALibXCPol():
    from deepchem.utils.dft_utils import CalcGGALibXCPol
    libxcfcn = pylibxc.LibXCFunctional("gga_c_pbe", "polarized")
    rho_u = torch.tensor([0.1, 0.2, 0.3])
    rho_d = torch.tensor([0.1, 0.2, 0.3])
    sigma_uu = torch.tensor([0.1, 0.2, 0.3])
    sigma_ud = torch.tensor([0.1, 0.2, 0.3])
    sigma_dd = torch.tensor([0.1, 0.2, 0.3])
    res = CalcGGALibXCPol.apply(rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd, 0,
                                libxcfcn)[0]
    assert torch.allclose(
        res, torch.tensor([[-0.0047, -0.0175, -0.0322]], dtype=torch.float64),
        0.0001, 0.0001)


@pytest.mark.torch
def test_CalcMGGALibXCUnpol():
    from deepchem.utils.dft_utils import CalcMGGALibXCUnpol
    libxcfcn = pylibxc.LibXCFunctional("mgga_c_m06_l", "unpolarized")
    rho = torch.tensor([0.1, 0.2, 0.3])
    sigma = torch.tensor([0.1, 0.2, 0.3])
    lapl = torch.tensor([0.1, 0.2, 0.3])
    kin = torch.tensor([0.1, 0.2, 0.3])
    res = CalcMGGALibXCUnpol.apply(rho, sigma, lapl, kin, 0, libxcfcn)[0]
    assert torch.allclose(
        res, torch.tensor([[-0.0032, -0.0066, -0.0087]], dtype=torch.float64),
        0.0001, 0.0001)


@pytest.mark.torch
def test_CalcMGGALibXCPol():
    from deepchem.utils.dft_utils import CalcMGGALibXCPol
    libxcfcn = pylibxc.LibXCFunctional("mgga_c_m06_l", "polarized")
    rho_u = torch.tensor([0.1, 0.2, 0.3])
    rho_d = torch.tensor([0.1, 0.2, 0.3])
    sigma_uu = torch.tensor([0.1, 0.2, 0.3])
    sigma_ud = torch.tensor([0.1, 0.2, 0.3])
    sigma_dd = torch.tensor([0.1, 0.2, 0.3])
    lapl_u = torch.tensor([0.1, 0.2, 0.3])
    lapl_d = torch.tensor([0.1, 0.2, 0.3])
    kin_u = torch.tensor([0.1, 0.2, 0.3])
    kin_d = torch.tensor([0.1, 0.2, 0.3])
    res = CalcMGGALibXCPol.apply(rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd,
                                 lapl_u, lapl_d, kin_u, kin_d, 0, libxcfcn)[0]
    assert torch.allclose(
        res, torch.tensor([[-0.0065, -0.0115, -0.0162]], dtype=torch.float64),
        0.0001, 0.0001)


@pytest.mark.torch
def test_get_libxc_res():
    from deepchem.utils.dft_utils.xc.libxc_wrapper import _get_libxc_res
    libxcfcn = pylibxc.LibXCFunctional("lda_x", "unpolarized")
    rho = torch.tensor([0.1, 0.2, 0.3])
    res = _get_libxc_res({"rho": rho}, 0, libxcfcn, 2, False)
    assert torch.allclose(
        res[0], torch.tensor([[-0.0343, -0.0864, -0.1483]],
                             dtype=torch.float64), 0.0001, 0.0001)


@pytest.mark.torch
def test_pack_input():
    from deepchem.utils.dft_utils.xc.libxc_wrapper import _pack_input
    rho = torch.tensor([[1, 2], [3, 4]])
    sigma = torch.tensor([[1, 2], [3, 4]])
    assert np.allclose(_pack_input(rho, sigma),
                       np.array([[[1, 1], [3, 3]], [[2, 2], [4, 4]]]))


@pytest.mark.torch
def test_unpack_input():
    from deepchem.utils.dft_utils.xc.libxc_wrapper import _unpack_input
    inp = np.array([[1, 3], [2, 4]])
    assert np.allclose(tuple(_unpack_input(inp))[0], np.array([1, 2]))
    assert np.allclose(tuple(_unpack_input(inp))[1], np.array([3, 4]))


@pytest.mark.torch
def test_get_dos():
    from deepchem.utils.dft_utils.xc.libxc_wrapper import _get_dos
    assert _get_dos(0) == (True, False, False, False, False)


@pytest.mark.torch
def test_extract_returns():
    from deepchem.utils.dft_utils.xc.libxc_wrapper import _extract_returns
    ret = {"zk": np.array([1, 2, 3])}
    assert torch.allclose(
        _extract_returns(ret, 0, 1)[0], torch.tensor([1, 2, 3]))


@pytest.mark.torch
def test_get_grad_inps():
    from deepchem.utils.dft_utils.xc.libxc_wrapper import _get_grad_inps
    grad_res = [torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])]
    inps = [torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])]
    derivs = [torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])]
    needs_input_grad = [True, True]
    deriv_idxs = [[0], [1]]
    result = _get_grad_inps(grad_res, inps, derivs, needs_input_grad,
                            deriv_idxs)
    assert torch.allclose(result[0], torch.tensor([1, 4, 9]))
    assert torch.allclose(result[1], torch.tensor([4, 10, 18]))


@pytest.mark.torch
def test_LibXCLDA():
    from deepchem.utils.dft_utils import ValGrad, LibXCLDA
    # create a LDA wrapper for libxc
    lda = LibXCLDA("lda_x")
    # create a density information
    densinfo = ValGrad(value=torch.rand(2, 3, 4), grad=torch.rand(2, 3, 4, 3))
    # get the exchange-correlation potential
    potinfo = lda.get_vxc(densinfo)
    assert potinfo.value.shape == torch.Size([2, 3, 4])
    edens = lda.get_edensityxc(densinfo)
    assert edens.shape == torch.Size([2, 3, 4])


@pytest.mark.torch
def test_LibXCGGA():
    from deepchem.utils.dft_utils import ValGrad, LibXCGGA
    # create a GGA wrapper for libxc
    gga = LibXCGGA("gga_c_pbe")
    # create a density information
    n = 2
    rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    densinfo = ValGrad(value=rho_u, grad=grad_u)
    # get the exchange-correlation potential
    potinfo = gga.get_vxc(densinfo)
    assert potinfo.value.shape == torch.Size([2])


@pytest.mark.torch
def test_LibXCMGGA():
    from deepchem.utils.dft_utils import ValGrad, LibXCMGGA
    # create a MGGA wrapper for libxc
    mgga = LibXCMGGA("mgga_x_scan")
    # create a density information
    n = 2
    rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    lapl_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    kin_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    densinfo = ValGrad(value=rho_u, grad=grad_u, lapl=lapl_u, kin=kin_u)
    # get the exchange-correlation potential
    potinfo = mgga.get_vxc(densinfo)
    assert potinfo.value.shape == torch.Size([2])


@pytest.mark.torch
def test_prepare_libxc_input():
    from deepchem.utils.dft_utils import ValGrad, SpinParam
    from deepchem.utils.dft_utils.xc.libxc import _prepare_libxc_input
    # create a density information
    n = 2
    rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    lapl_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    kin_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    rho_d = torch.rand((n,), dtype=torch.float64).requires_grad_()
    grad_d = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    lapl_d = torch.rand((n,), dtype=torch.float64).requires_grad_()
    kin_d = torch.rand((n,), dtype=torch.float64).requires_grad_()
    densinfo = SpinParam(u=ValGrad(value=rho_u,
                                   grad=grad_u,
                                   lapl=lapl_u,
                                   kin=kin_u),
                         d=ValGrad(value=rho_d,
                                   grad=grad_d,
                                   lapl=lapl_d,
                                   kin=kin_d))
    # prepare the input for libxc
    inputs = _prepare_libxc_input(densinfo, 4)
    assert len(inputs) == 9


@pytest.mark.torch
def test_postproc_libxc_voutput():
    from deepchem.utils.dft_utils import ValGrad
    from deepchem.utils.dft_utils.xc.libxc import _postproc_libxc_voutput
    # create a density information
    n = 2
    rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    lapl_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    kin_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    densinfo = ValGrad(value=rho_u, grad=grad_u, lapl=lapl_u, kin=kin_u)
    # postprocess the output from libxc
    potinfo = _postproc_libxc_voutput(
        densinfo,
        torch.rand((n,), dtype=torch.float64).requires_grad_())
    assert potinfo.value.shape == torch.Size([2])


@pytest.mark.torch
def test_get_libxc():
    from deepchem.utils.dft_utils import ValGrad, get_libxc
    xc = get_libxc("gga_c_pbe")
    n = 2
    rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    densinfo = ValGrad(value=rho_u, grad=grad_u)
    # get the exchange-correlation potential
    potinfo = xc.get_vxc(densinfo)
    assert potinfo.value.shape == torch.Size([2])


@pytest.mark.torch
def test_get_xc():
    from deepchem.utils.dft_utils import ValGrad, get_xc
    xc = get_xc("lda_x + gga_c_pbe")
    n = 2
    rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    densinfo = ValGrad(value=rho_u, grad=grad_u)
    # get the exchange-correlation potential
    potinfo = xc.get_vxc(densinfo)
    assert potinfo.value.shape == torch.Size([2])


@pytest.mark.torch
def test_libcintwrapper():
    from deepchem.utils.dft_utils import AtomCGTOBasis, LibcintWrapper, loadbasis
    dtype = torch.double
    d = 1.0
    pos_requires_grad = True
    pos1 = torch.tensor([0.1 * d, 0.0 * d, 0.2 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    pos2 = torch.tensor([0.0 * d, 1.0 * d, -0.4 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    poss = [pos1, pos2, pos3]
    atomzs = [1, 1, 1]

    allbases = [
        loadbasis("%d:%s" % (max(atomz, 1), "3-21G"),
                  dtype=dtype,
                  requires_grad=False) for atomz in atomzs
    ]

    atombases = [
        AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i]) \
        for i in range(len(allbases))
    ]
    wrap = LibcintWrapper(atombases, True, None)
    assert wrap.ao_idxs() == (0, 6)


@pytest.mark.torch
def test_SubsetLibcintWrapper():
    from deepchem.utils.dft_utils import AtomCGTOBasis, LibcintWrapper, loadbasis
    dtype = torch.double
    d = 1.0
    pos_requires_grad = True
    pos1 = torch.tensor([0.1 * d, 0.0 * d, 0.2 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    pos2 = torch.tensor([0.0 * d, 1.0 * d, -0.4 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    poss = [pos1, pos2, pos3]
    atomzs = [1, 1, 1]
    allbases = [
        loadbasis("%d:%s" % (max(atomz, 1), "3-21G"),
                  dtype=dtype,
                  requires_grad=False) for atomz in atomzs
    ]
    atombases = [
        AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
        for i in range(len(allbases))
    ]
    wrap = LibcintWrapper(atombases, True, None)
    subset = wrap[1:3]
    assert subset.ao_idxs() == (1, 3)


@pytest.mark.torch
def test_molintor():
    from deepchem.utils.dft_utils import AtomCGTOBasis, LibcintWrapper, loadbasis, \
        int1e, int2e, int2c2e, int3c2e, overlap, kinetic, nuclattr, elrep, coul2c, coul3c
    dtype = torch.double
    d = 1.0
    pos_requires_grad = True
    pos1 = torch.tensor([0.1 * d, 0.0 * d, 0.2 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    pos2 = torch.tensor([0.0 * d, 1.0 * d, -0.4 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d],
                        dtype=dtype,
                        requires_grad=pos_requires_grad)
    poss = [pos1, pos2, pos3]
    atomzs = [1, 1, 1]

    allbases = [
        loadbasis("%d:%s" % (max(atomz, 1), "3-21G"),
                  dtype=dtype,
                  requires_grad=False) for atomz in atomzs
    ]

    atombases = [
        AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i]) \
        for i in range(len(allbases))
    ]
    env = LibcintWrapper(atombases, True, None)

    assert int1e("r0", env).shape == torch.Size([3, 6, 6])
    assert int1e("r0r0", env).shape == torch.Size([9, 6, 6])
    assert int1e("r0r0r0", env).shape == torch.Size([27, 6, 6])
    assert int2e("ar12b", env).shape == torch.Size([6, 6, 6, 6])
    assert int2c2e("ipip1", env).shape == torch.Size([3, 3, 6, 6])
    assert int3c2e("ar12", env).shape == torch.Size([6, 6, 6])
    assert overlap(env).shape == torch.Size([6, 6])
    assert kinetic(env).shape == torch.Size([6, 6])
    assert nuclattr(env).shape == torch.Size([6, 6])
    assert elrep(env).shape == torch.Size([6, 6, 6, 6])
    assert coul2c(env).shape == torch.Size([6, 6])
    assert coul3c(env).shape == torch.Size([6, 6, 6])


@pytest.mark.torch
def test_intor_name_manager():
    from deepchem.utils.dft_utils.hamilton.intor.namemgr import IntorNameManager
    mgr = IntorNameManager("int1e", "r0")
    assert mgr.fullname == "int1e_r0"
    assert mgr.get_intgl_name(True) == "int1e_r0_sph"
    assert mgr.get_ft_intgl_name(True) == "GTO_ft_r0_sph"
    assert mgr.get_intgl_symmetry([0, 1, 2, 0]).code == "s1"
    assert mgr.get_intgl_components_shape() == (3,)


@pytest.mark.torch
def test_base_symmetry():
    from deepchem.utils.dft_utils.hamilton.intor.symmetry import BaseSymmetry

    class SemNew(BaseSymmetry):

        def get_reduced_shape(self, orig_shape):
            return orig_shape

        @property
        def code(self) -> str:
            return "sn"

        def reconstruct_array(self, arr, orig_shape):
            return arr

    sym = SemNew()
    assert sym.get_reduced_shape((2, 3, 4)) == (2, 3, 4)
    assert sym.code == 'sn'
    assert sym.reconstruct_array(torch.rand((2, 3, 4)),
                                 (2, 3, 4)).shape == torch.Size([2, 3, 4])


@pytest.mark.torch
def test_s1_symmetry():
    from deepchem.utils.dft_utils.hamilton.intor.symmetry import S1Symmetry
    sym = S1Symmetry()
    assert sym.get_reduced_shape((2, 3, 4)) == (2, 3, 4)
    assert sym.code == 's1'
    assert sym.reconstruct_array(np.random.rand(2, 3, 4),
                                 (2, 3, 4)).shape == torch.Size([2, 3, 4])


@pytest.mark.torch
def test_s4_symmetry():
    from deepchem.utils.dft_utils.hamilton.intor.symmetry import S4Symmetry
    sym = S4Symmetry()
    assert sym.get_reduced_shape((3, 3, 4, 4)) == (6, 10)
    assert sym.code == 's4'
    assert sym.reconstruct_array(np.random.rand(2, 3, 4, 4),
                                 (3, 3, 4, 4)).shape == (3, 3, 4, 4)


@pytest.mark.torch
def test_np2ctypes():
    """Just checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils.hamilton.intor.utils import np2ctypes
    arr = np.random.rand(2, 3, 4)
    np2ctypes(arr)


@pytest.mark.torch
def test_int2ctypes():
    """Just checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils.hamilton.intor.utils import int2ctypes
    arr = 51
    int2ctypes(arr)


@pytest.mark.torch
def test_memoize_method():
    from deepchem.utils import memoize_method

    class A:

        @memoize_method
        def foo(self):
            print("foo")
            return 1

    a = A()
    assert a.foo() == 1


@pytest.mark.torch
def test_load_basis():
    from deepchem.utils.dft_utils import loadbasis
    H = loadbasis("1:3-21G")
    assert H[0].alphas.shape == torch.Size([2])


@pytest.mark.torch
def test_ks_engine():
    """Tests KSEngine and KS Class."""
    from deepchem.utils.dft_utils import (BaseHamilton, BaseSystem, BaseGrid,
                                          SpinParam, KSEngine)
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

        def _rmv(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _mm(self, x):
            return torch.matmul(self.param, x)

        def _rmm(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _fullmatrix(self):
            return self.param

    class MyHamilton(BaseHamilton):

        def __init__(self):
            self._nao = 2
            self._kpts = torch.tensor([[0.0, 0.0, 0.0]])
            self._df = None

        @property
        def nao(self):
            return self._nao

        @property
        def kpts(self):
            return self._kpts

        @property
        def df(self):
            return self._df

        def build(self):
            return self

        def get_nuclattr(self):
            return torch.ones((1, 1, self.nao, self.nao))

        def get_e_elrep(self, dmtot):
            return 2 * dmtot

        def get_e_exchange(self, dm):
            return 3 * dm

        def get_e_hcore(self, dm):
            return 4.0 * dm

        def get_elrep(self, dmtot):
            return MyLinOp((self.nao + 1, self.nao + 1))

        def get_exchange(self, dm):
            return MyLinOp((self.nao + 1, self.nao + 1))

        def get_kinnucl(self):
            linop = MyLinOp((self.nao + 1, self.nao + 1))
            return linop

        def ao_orb2dm(self, orb: torch.Tensor,
                      orb_weight: torch.Tensor) -> torch.Tensor:
            return orb * orb_weight

    ham = MyHamilton()

    class MySystem(BaseSystem):

        def __init__(self):
            self.hamiltonian = ham
            self.grid = BaseGrid()

        def get_hamiltonian(self):
            return self.hamiltonian

        def get_grid(self):
            return self.grid

        def requires_grid(self):
            return True

        def get_orbweight(
            self,
            polarized: bool = False
        ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
            return SpinParam(torch.tensor([1.0]), torch.tensor([2.0]))

        def get_nuclei_energy(self):
            return 10.0

    system = MySystem()
    engine = KSEngine(system, None)
    engine.set_eigen_options(eigen_options={"method": "exacteig"})

    assert engine.dm2energy(torch.tensor([2])) == torch.tensor([22.0])
    assert engine.dm2scp(torch.tensor([2])).shape == torch.Size([3, 3])
    assert engine.scp2dm(torch.rand((2, 2, 2))).u.shape == torch.Size([2, 1])


@pytest.mark.torch
def test_read_float():
    from deepchem.utils.dft_utils.api.loadbasis import _read_float
    assert _read_float("1.0D+00") == 1.0


@pytest.mark.torch
def test_get_basis_file():
    from deepchem.utils.dft_utils.api.loadbasis import _get_basis_file
    fname = _get_basis_file("1:3-21G")
    path = fname.split("/")[-1]
    assert path == "01.gaussian94"


@pytest.mark.torch
def test_normalize_basisname():
    from deepchem.utils.dft_utils.api.loadbasis import _normalize_basisname
    assert _normalize_basisname("6-311++G**") == '6-311ppgss'


@pytest.mark.torch
def test_expand_angmoms():
    from deepchem.utils.dft_utils.api.loadbasis import _expand_angmoms
    assert _expand_angmoms("SP", 2) == [0, 1]



"""Normalizing flows for transforming probability distributions using PyTorch.
"""

try:
    from deepchem.models.torch_models.flows import NormalizingFlow  # noqa
except ModuleNotFoundError:
    raise ImportError('These classes require PyTorch to be installed.')



"""
Clinical Toxicity (clintox) dataset loader.
@author Caleb Geniesse
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

CLINTOX_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/clintox.csv.gz"
CLINTOX_TASKS = ['FDA_APPROVED', 'CT_TOX']


class _ClintoxLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "clintox.csv.gz")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=CLINTOX_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_clintox(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['balancing'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load ClinTox dataset

    The ClinTox dataset compares drugs approved by the FDA and
    drugs that have failed clinical trials for toxicity reasons.
    The dataset includes two classification tasks for 1491 drug
    compounds with known chemical structures:

    #. clinical trial toxicity (or absence of toxicity)
    #. FDA approval status.

    List of FDA-approved drugs are compiled from the SWEETLEAD
    database, and list of drugs that failed clinical trials for
    toxicity reasons are compiled from the Aggregate Analysis of
    ClinicalTrials.gov(AACT) database.

    Random splitting is recommended for this dataset.

    The raw data csv file contains columns below:

    - "smiles" - SMILES representation of the molecular structure
    - "FDA_APPROVED" - FDA approval status
    - "CT_TOX" - Clinical trial results

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    References
    ----------
    .. [1] Gayvert, Kaitlyn M., Neel S. Madhukar, and Olivier Elemento.
        "A data-driven approach to predicting successes and failures of clinical
        trials."
        Cell chemical biology 23.10 (2016): 1294-1301.
    .. [2] Artemov, Artem V., et al. "Integrated deep learned transcriptomic and
        structure-based predictor of clinical trials outcomes." bioRxiv (2016):
        095653.
    .. [3] Novick, Paul A., et al. "SWEETLEAD: an in silico database of approved
        drugs, regulated chemicals, and herbal isolates for computer-aided drug
        discovery." PloS one 8.11 (2013): e79568.
    .. [4] Aggregate Analysis of ClincalTrials.gov (AACT) Database.
        https://www.ctti-clinicaltrials.org/aact-database
    """
    loader = _ClintoxLoader(featurizer, splitter, transformers, CLINTOX_TASKS,
                            data_dir, save_dir, **kwargs)
    return loader.load_dataset('clintox', reload)



"""
Making it easy to import in classes.
"""
# flake8: noqa
import logging

logger = logging.getLogger(__name__)

# base classes for featurizers
from deepchem.feat.base_classes import Featurizer
from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.feat.base_classes import MaterialStructureFeaturizer
from deepchem.feat.base_classes import MaterialCompositionFeaturizer
from deepchem.feat.base_classes import ComplexFeaturizer
from deepchem.feat.base_classes import UserDefinedFeaturizer
from deepchem.feat.base_classes import DummyFeaturizer
from deepchem.feat.base_classes import PolymerFeaturizer

from deepchem.feat.graph_features import ConvMolFeaturizer
from deepchem.feat.graph_features import WeaveFeaturizer
from deepchem.feat.graph_data import GraphData
from deepchem.feat.binding_pocket_features import BindingPocketFeaturizer

# molecule featurizers
from deepchem.feat.molecule_featurizers import AtomicCoordinates
from deepchem.feat.molecule_featurizers import BPSymmetryFunctionInput
from deepchem.feat.molecule_featurizers import CircularFingerprint
from deepchem.feat.molecule_featurizers import CoulombMatrix
from deepchem.feat.molecule_featurizers import CoulombMatrixEig
from deepchem.feat.molecule_featurizers import MACCSKeysFingerprint
from deepchem.feat.molecule_featurizers import MordredDescriptors
from deepchem.feat.molecule_featurizers import Mol2VecFingerprint
from deepchem.feat.molecule_featurizers import MolGraphConvFeaturizer
from deepchem.feat.molecule_featurizers import PagtnMolGraphFeaturizer
from deepchem.feat.molecule_featurizers import MolGanFeaturizer
from deepchem.feat.molecule_featurizers import OneHotFeaturizer
from deepchem.feat.molecule_featurizers import SparseMatrixOneHotFeaturizer
from deepchem.feat.molecule_featurizers import PubChemFingerprint
from deepchem.feat.molecule_featurizers import RawFeaturizer
from deepchem.feat.molecule_featurizers import RDKitDescriptors
from deepchem.feat.molecule_featurizers import SmilesToImage
from deepchem.feat.molecule_featurizers import SmilesToSeq, create_char_to_idx
from deepchem.feat.molecule_featurizers import MATFeaturizer
from deepchem.feat.molecule_featurizers import DMPNNFeaturizer
from deepchem.feat.molecule_featurizers import GroverFeaturizer
from deepchem.feat.molecule_featurizers import SNAPFeaturizer
from deepchem.feat.molecule_featurizers import RDKitConformerFeaturizer
from deepchem.feat.molecule_featurizers import MXMNetFeaturizer

# complex featurizers
from deepchem.feat.complex_featurizers import RdkitGridFeaturizer
from deepchem.feat.complex_featurizers import NeighborListAtomicCoordinates
from deepchem.feat.complex_featurizers import NeighborListComplexAtomicCoordinates
from deepchem.feat.complex_featurizers import AtomicConvFeaturizer
from deepchem.feat.complex_featurizers import ComplexNeighborListFragmentAtomicCoordinates
from deepchem.feat.complex_featurizers import ContactCircularFingerprint
from deepchem.feat.complex_featurizers import ContactCircularVoxelizer
from deepchem.feat.complex_featurizers import SplifFingerprint
from deepchem.feat.complex_featurizers import SplifVoxelizer
from deepchem.feat.complex_featurizers import ChargeVoxelizer
from deepchem.feat.complex_featurizers import SaltBridgeVoxelizer
from deepchem.feat.complex_featurizers import CationPiVoxelizer
from deepchem.feat.complex_featurizers import PiStackVoxelizer
from deepchem.feat.complex_featurizers import HydrogenBondVoxelizer
from deepchem.feat.complex_featurizers import HydrogenBondCounter

# material featurizers
from deepchem.feat.material_featurizers import ElementPropertyFingerprint
from deepchem.feat.material_featurizers import SineCoulombMatrix
from deepchem.feat.material_featurizers import CGCNNFeaturizer
from deepchem.feat.material_featurizers import ElemNetFeaturizer
from deepchem.feat.material_featurizers import LCNNFeaturizer

from deepchem.feat.atomic_conformation import AtomicConformation
from deepchem.feat.atomic_conformation import AtomicConformationFeaturizer

from deepchem.feat.huggingface_featurizer import HuggingFaceFeaturizer

# biological sequence featurizers
try:
    from deepchem.feat.bio_seq_featurizer import SAMFeaturizer
    from deepchem.feat.bio_seq_featurizer import BAMFeaturizer
    from deepchem.feat.bio_seq_featurizer import CRAMFeaturizer
except ImportError as e:
    logger.warning(
        f'Skipped loading biological sequence featurized, missing a dependency. {e}'
    )

# tokenizers
try:
    from deepchem.feat.smiles_tokenizer import SmilesTokenizer
    from deepchem.feat.smiles_tokenizer import BasicSmilesTokenizer
    from deepchem.feat.bert_tokenizer import BertFeaturizer
    from deepchem.feat.roberta_tokenizer import RobertaFeaturizer
    from deepchem.feat.reaction_featurizer import RxnFeaturizer
except ModuleNotFoundError:
    pass

from deepchem.feat.vocabulary_builders import HuggingFaceVocabularyBuilder

# support classes
from deepchem.feat.molecule_featurizers import GraphMatrix
